# game.py

from pygame.locals import *
import framework, pygame, random

# distances
NEAR, MIDDLE, FAR = 0, 1, 2
MAX_BULLETS = 5
NUM_FURBIES_START = 10
SPAWNS, CORPSES = 0, 1
SCORE_MSG_DUR = 600
SCORE_MSG_MOV_SPEED = 2

class Game:	
	class VisObject:
		def __init__(self, x, y):
			self.pos = [x, y]
	
	class FadeOutSequence(VisObject):
		def __init__(self, x, y, image, speed, ignoreScroll = False):
			Game.VisObject.__init__(self, x, y)
			self.speed = speed
			self.image = image
			self.alpha = 255
			self.ignoreScroll = ignoreScroll
		
		def update(self):
			self.alpha -= self.speed
			return self.alpha <= 0
		
		def draw(self, scrollX):			
			x = self.pos[0] - scrollX
			
			if x + self.image.get_width() < 0 or x > framework.screen.get_width() and not self.ignoreScroll:
				return
			
			self.image.set_alpha(self.alpha)
			
			if not self.ignoreScroll:
				framework.draw(self.image, x, self.pos[1])
			else:
				framework.draw(self.image, self.pos[0], self.pos[1])
			
			self.image.set_alpha(None)
	
	class ScoreMsg(VisObject):
		def __init__(self, x, y, score, font, color):
			Game.VisObject.__init__(self, x, y)
			self.score = score
			self.color = color
			self.startTime = pygame.time.get_ticks()
			self.font = font
		
		def update(self):
			self.pos[1] -= SCORE_MSG_MOV_SPEED
			return pygame.time.get_ticks() - self.startTime > SCORE_MSG_DUR or self.pos[0] < 0
		
		def draw(self, scrollX):
			x = self.pos[0] - scrollX
			framework.drawText(self.font, "+ " + str(self.score), self.color, x, self.pos[1])
	
	class Furby(VisObject):
		def __init__(self, x, y, distance, imgW, imgH):
			Game.VisObject.__init__(self, x, y)
			
			self.outOfBorder = False
			self.distance = distance
			
			if distance == NEAR:
				self.movSpeed = 4
				self.score = 100
			elif distance == MIDDLE:
				self.movSpeed = 3
				self.score = 150
			else:
				self.movSpeed = 2
				self.score = 300
			
			self.imgW = imgW
			self.imgH = imgH
		
		def update(self, areaWidth):
			self.pos[0] += self.movSpeed			
			self.outOfBorder = self.pos[0] - self.imgW < 0 or \
				self.pos[0] > areaWidth
			
		def isVisible(self, scrollX):
			return not (self.pos[0] + self.imgW < scrollX or self.pos[0] > scrollX + framework.screen.get_width())
		
		def getDrawnRect(self, scrollX):
			return Rect(self.pos[0] - scrollX, self.pos[1], self.imgW, self.imgH)
		
		def draw(self, furbyImgs, scrollX):
			# Not inside the screen, don't draw it!
			if not self.isVisible(scrollX):
				return
											
			x = self.pos[0] - scrollX			
			framework.draw(furbyImgs[self.distance], x, self.pos[1])
	
	class StaticObj(VisObject):
		def __init__(self, x, y, distance, images, duration = -1, isFalling = False):
			Game.VisObject.__init__(self, x, y)
			self.image = images[distance]
			self.startTime = pygame.time.get_ticks()
			self.duration = duration
			self.isFalling = isFalling
			self.distance = distance
		
		def update(self, fallSpeed = -1):
			if self.isFalling:
				return self.updateFalling(fallSpeed)
			else:
				if self.duration == -1:
					return False
				else:
					return (pygame.time.get_ticks() - self.startTime > self.duration)
		
		def updateFalling(self, fallSpeed):
			self.pos[1] += fallSpeed
			return self.pos[1] > framework.screen.get_height()
		
		def isVisible(self, scrollX):
			return self.pos[0] + self.image.get_width() > scrollX and \
					self.pos[0] < scrollX + framework.screen.get_width()
		
		def draw(self, scrollX):
			if not self.isVisible(scrollX):
				return
			
			x = self.pos[0] - scrollX
			framework.draw(self.image, x, self.pos[1])
	
	def init(self):
		self.done = False
		
		# Load assets
		self.crosshairImg = framework.loadImage("crosshair.bmp")
		self.backgroundImg = framework.loadImage("landscape.jpg", False)
		
		self.furbyImgs = []
		self.furbyImgs.append(framework.loadImage("enemy_near.bmp"))
		self.furbyImgs.append(framework.loadImage("enemy_middle.bmp"))
		self.furbyImgs.append(framework.loadImage("enemy_far.bmp"))
		
		self.corpseImgs = []
		self.corpseImgs.append(framework.loadImage("corpse_near.bmp"))
		self.corpseImgs.append(framework.loadImage("corpse_middle.bmp"))
		self.corpseImgs.append(framework.loadImage("corpse_far.bmp"))
		
		self.spawnImgs = []
		self.spawnImgs.append(framework.loadImage("spawn_near.bmp"))
		self.spawnImgs.append(framework.loadImage("spawn_middle.bmp"))
		self.spawnImgs.append(framework.loadImage("spawn_far.bmp"))
		
		self.bulletImg = framework.loadImage("bullet.bmp")
		
		framework.loadAndPlayMusic("music.ogg")
		
		# Vars needed for scrolling
		self.scrollX = (self.backgroundImg.get_width() - framework.screen.get_width()) / 2
		self.scrollSpeed = 8
		self.leftScrlBorder = framework.screen.get_width() / 6
		self.rightScrlBorder = framework.screen.get_width() - self.leftScrlBorder
		
		# Center crosshair
		centerPos = (framework.screen.get_width() / 2, framework.screen.get_height() / 2)
		pygame.mouse.set_pos(centerPos)
		self.cX = centerPos[0] - self.crosshairImg.get_width() / 2
		self.cY = centerPos[1] - self.crosshairImg.get_height() / 2
		
		self.furbySpawnTime = 1000
		self.lFurbySpawn = pygame.time.get_ticks()
		random.seed()		
		
		self.nearFurbies, self.middleFurbies, self.farFurbies = [], [], []
		self.furbyLists = [ self.farFurbies, self.middleFurbies, self.nearFurbies ]
		
		self.nearStaticObjs, self.middleStaticObjs, self.farStaticObjs = [], [], []
		self.staticObjs = [ self.farStaticObjs, self.middleStaticObjs, self.nearStaticObjs ]
		
		self.roundStarted = pygame.time.get_ticks()
		self.hudFont = pygame.font.SysFont("arial", 24)
		self.score = 0
		self.roundTime = 120000
		self.bulletsInMag = MAX_BULLETS
		self.mbtnWasReleased = False
		
		self.shotSnd = framework.loadSound("shot.ogg")
		self.reloadSnd = framework.loadSound("reload.ogg")
		self.screamSnd = framework.loadSound("scream.ogg")
		self.startRoundSnd = framework.loadSound("start_round.ogg")
		
		self.startRoundSnd.play()
		self.gameOver = False
		
		self.fallSpeeds = (16, 8, 4)
		
		self.scoreMsgs = []
		self.scoreFont = pygame.font.SysFont("arial", 32)
		self.scoreCol = (0, 255, 0)
		
		for i in range(0, NUM_FURBIES_START):
			self.spawnFurby()
		
		self.fadeoutSeqs = []
	
	def input(self):
		self.done = framework.keyPressed(K_ESCAPE)
		self.mousePos = pygame.mouse.get_pos()
		
		# Scrolling		
		if framework.keyPressed(K_LEFT, False) or self.mousePos[0] < self.leftScrlBorder:
			self.scrollX -= self.scrollSpeed
		
		if framework.keyPressed(K_RIGHT, False) or self.mousePos[0] > self.rightScrlBorder:
			self.scrollX += self.scrollSpeed
				
		if self.scrollX < 0:
			self.scrollX = 0
		
		if self.scrollX + framework.screen.get_width() > self.backgroundImg.get_width():
			self.scrollX = self.backgroundImg.get_width() - framework.screen.get_width()
		
		if pygame.mouse.get_pressed()[0]:
			self.shoot()
		else:
			self.wasReleased = True
		
		if pygame.mouse.get_pressed()[2]:
			self.bulletsInMag = MAX_BULLETS
			self.reloadSnd.play()
	
	def addCorpse(self, pos, distance):
		newCorpse = Game.StaticObj(pos[0], pos[1], distance, self.corpseImgs, -1, True)
		if distance == NEAR:
			self.nearStaticObjs.append(newCorpse)
		elif distance == MIDDLE:
			self.middleStaticObjs.append(newCorpse)
		else:
			self.farStaticObjs.append(newCorpse)
	
	def shoot(self):
		if self.wasReleased and not self.bulletsInMag < 1:
			#self.crosshairRect = Rect(self.cX, self.cY, \
			#						  self.crosshairImg.get_width(), \
			#						  self.crosshairImg.get_height())
			
			for furbyList in self.furbyLists:
				for furby in furbyList:
					if furby.isVisible(self.scrollX) and \
						framework.pointColl(furby.getDrawnRect(self.scrollX), self.mousePos):
						self.score += furby.score
						scoreX = furby.pos[0]
						scoreY = furby.pos[1] + furby.imgH / 2
						self.scoreMsgs.append(Game.ScoreMsg(scoreX, scoreY, furby.score, self.scoreFont, self.scoreCol))
						self.addCorpse(furby.pos, furby.distance)
						self.fadeoutSeqs.append(Game.FadeOutSequence(furby.pos[0], furby.pos[1], \
																	 self.furbyImgs[furby.distance], 50))
						furbyList.remove(furby)
						self.screamSnd.play()
			
			self.bulletsInMag -= 1
			
			foX = framework.screen.get_width() - (MAX_BULLETS - self.bulletsInMag) * self.bulletImg.get_width()
			foY = framework.screen.get_height() - self.bulletImg.get_height()
			self.fadeoutSeqs.append(Game.FadeOutSequence(foX, foY, self.bulletImg, 25, True))
			
			self.wasReleased = False
			self.shotSnd.play()
	
	def spawnFurby(self):
		itsX = random.randint(0, self.backgroundImg.get_width() - self.furbyImgs[0].get_width())
		itsY = random.randint(0, self.backgroundImg.get_height() - self.furbyImgs[0].get_height())
		
		distance = random.randint(0, 2)
		
		newFurby = Game.Furby(itsX, itsY, distance, \
							  self.furbyImgs[distance].get_width(), \
							  self.furbyImgs[distance].get_height())
		
		statObj = Game.StaticObj(itsX, itsY, distance, \
								 self.spawnImgs, 200)
		
		if distance == NEAR:
			self.nearFurbies.append(newFurby)
			self.nearStaticObjs.append(statObj)
		elif distance == MIDDLE:
			self.middleFurbies.append(newFurby)
			self.middleStaticObjs.append(statObj)
		else:
			self.farFurbies.append(newFurby)
			self.farStaticObjs.append(statObj)
		
		self.lFurbySpawn = pygame.time.get_ticks()
	
	def update(self):
		if self.gameOver:
			return
		
		# Spawn furby
		if pygame.time.get_ticks() - self.lFurbySpawn > self.furbySpawnTime:
			self.spawnFurby()
		
		for furbyList in self.furbyLists:
			for furby in furbyList:
				furby.update(self.backgroundImg.get_width())
				if furby.outOfBorder:
					furbyList.remove(furby)
		
		for statObjList in self.staticObjs:
			for staticObj in statObjList:
				fallSpeed = self.fallSpeeds[staticObj.distance]
				result = staticObj.update(fallSpeed)
				if result:
					statObjList.remove(staticObj)
		
		for scoreMsg in self.scoreMsgs:
			result = scoreMsg.update()
			if result:
				self.scoreMsgs.remove(scoreMsg)
		
		for fadeoutSeq in self.fadeoutSeqs:
			result = fadeoutSeq.update()
			if result:
				self.fadeoutSeqs.remove(fadeoutSeq)
		
		if pygame.time.get_ticks() - self.roundStarted > self.roundTime and not self.gameOver:
			self.gameOver = True
	
	def drawBackground(self):
		srcRect = Rect(self.scrollX, 0, framework.screen.get_width(), \
					   framework.screen.get_height())
		destRect = Rect(0, 0, framework.screen.get_width(), \
						framework.screen.get_height())
		framework.screen.blit(self.backgroundImg, destRect, srcRect)
	
	def drawFurbies(self):
		for furbyList in self.furbyLists:
			for furby in furbyList:
				furby.draw(self.furbyImgs, self.scrollX)
	
	def drawStatObjs(self, type):
		for statObjList in self.staticObjs:
			for staticObj in statObjList:
				if type == SPAWNS and not staticObj.isFalling \
					or type == CORPSES and staticObj.isFalling:
					staticObj.draw(self.scrollX)
	
	def drawScoreMsgs(self):
		for scoreMsg in self.scoreMsgs:
			scoreMsg.draw(self.scrollX)
	
	def drawFadeOutSequences(self):
		for fadeOutSeq in self.fadeoutSeqs:
			fadeOutSeq.draw(self.scrollX)
	
	def drawCrosshair(self):
		self.cX = self.mousePos[0] - self.crosshairImg.get_width() / 2
		self.cY = self.mousePos[1] - self.crosshairImg.get_height() / 2
		framework.draw(self.crosshairImg, self.cX, self.cY)
	
	def drawHUD(self):                
		timeLeft = self.roundTime - (pygame.time.get_ticks() - self.roundStarted)
		timeLeft /= 1000
		mins = int(timeLeft / 60)
		secs = timeLeft - mins * 60
		
		if secs <= 9:
			secsStr = "0" + str(secs)
		else:
			secsStr = str(secs)
		
		framework.drawText(self.hudFont, "Score: %s" % self.score, framework.black, 100, 0)
		framework.drawText(self.hudFont, "Time left: %d:%s" % (mins, secsStr), framework.black, 400, 0)
		x = framework.screen.get_width() - self.bulletImg.get_width() * MAX_BULLETS
		for i in range(self.bulletsInMag):
			framework.draw(self.bulletImg, x, framework.screen.get_height() - self.bulletImg.get_height())
			x += self.bulletImg.get_width()
	
	def drawCenteredScore(self):
		framework.drawTextCentered(self.scoreFont, "Score: " + str(self.score), framework.black)
	
	def draw(self):
		self.drawBackground()
		self.drawStatObjs(CORPSES)
		self.drawFurbies()
		self.drawStatObjs(SPAWNS)
		self.drawScoreMsgs()
		self.drawFadeOutSequences()
		self.drawCrosshair()		
		
		if self.gameOver:
			self.drawCenteredScore()
		else:
			self.drawHUD()
